@@ -15,14 +15,16 @@ address. The list of telemetry items can be serialized after the fact and loaded
 lengthy discovery process.
 """
 import time
-from typing import Union, Optional
+from typing import Union, Optional, List, Any
 
 from .i2c import SupMCUMaster
-from .parsing import parse_telemetry
-from .types import SupMCUTelemetryDefinition, SupMCUModuleDefinition, TelemetryType
+from .parsing import parse_telemetry, _format_to_length
+from .types import SupMCUTelemetryDefinition, SupMCUModuleDefinition, TelemetryType, \
+    _supmcu_fmt_data_type, _data_type_sizes, TelemetryDataItem
 from ..i2c import I2CMaster
 
 SIZEOF_HEADER_FOOTER = 13
+SUPMCU_VERSION_DEFINITION = SupMCUTelemetryDefinition("Firmware Version", 77, 0, "S")
 SUPMCU_LENGTH_DEFINITION = SupMCUTelemetryDefinition("Length", 2 + SIZEOF_HEADER_FOOTER, 0, "s")
 SUPMCU_NAME_DEFINITION = SupMCUTelemetryDefinition("Name", 33 + SIZEOF_HEADER_FOOTER, 0, "S")
 SUPMCU_FORMAT_DEFINITION = SupMCUTelemetryDefinition("Format", 25 + SIZEOF_HEADER_FOOTER, 0, "S")
@@ -52,11 +54,36 @@ def _write_read_supcmu_i2c(i2c_master: I2CMaster,
     return i2c_master.read(address, read_length)
 
 
+def get_version_string(i2c_master: I2CMaster, address: int) -> str:
+    module_def = SupMCUModuleDefinition("", "", address, {0: SUPMCU_VERSION_DEFINITION}, {})
+    master = SupMCUMaster(i2c_master, (module_def,))
+    version = master.request_telemetry(address, TelemetryType.SupMCU, 0).items[0].value
+    # print(version)
+    return version
+
+
+def get_values(i2c_master: I2CMaster, address: int, module_cmd_name: str, idx: int, format: str, response_delay: Optional[float] = None) -> List[TelemetryDataItem]:
+    try:
+        size = _format_to_length(format) + SIZEOF_HEADER_FOOTER
+    except ValueError:
+        # Size of version string
+        size = SUPMCU_VERSION_DEFINITION.telemetry_length
+    # size = SIZEOF_HEADER_FOOTER + sum((_data_type_sizes[f] for f in fmt))
+    print(size)
+    print(format)
+    raw_bytes = _write_read_supcmu_i2c(i2c_master, address, f'{module_cmd_name}:TEL? {idx}\n',
+        size, response_delay)
+    # raw_bytes = raw_bytes[SIZEOF_HEADER_FOOTER:]
+    vals = parse_telemetry(raw_bytes, SupMCUTelemetryDefinition('', size, idx, format))
+    return vals.items
+
+
 def request_telemetry_definition(i2c_master: Union[I2CMaster, SupMCUMaster],
                                  address: int,
                                  module_cmd_name: str,
                                  idx: int,
-                                 response_delay: Optional[float] = None) -> SupMCUTelemetryDefinition:
+                                 response_delay: Optional[float] = None,
+                                 is_simulatable_mod: bool = False) -> SupMCUTelemetryDefinition:
     """
     Requests the formatting, name and length information from the device at I2C address `address`, using the module
     short name `module_cmd_name` (e.g. BM for Battery Module), concatenating that with `idx` in a telemetry request
@@ -69,8 +96,10 @@ def request_telemetry_definition(i2c_master: Union[I2CMaster, SupMCUMaster],
     :param idx: The telemetry index to grab the information for.
     :param response_delay: The amount of time in seconds to wait between I2C Write and read. Can be None, or set
                             from SupMCUMaster passed in as `i2c_master`.
+    :param is_simulatable_mod: Whether or not the module is simulatable.
     :return: The SupMCUTelemetryDefinition that represents the Telemetry data.
     """
+    is_simulatable = False
     if isinstance(i2c_master, SupMCUMaster):
         if response_delay is None:
             response_delay = i2c_master.request_delay
@@ -86,12 +115,19 @@ def request_telemetry_definition(i2c_master: Union[I2CMaster, SupMCUMaster],
                                         SUPMCU_NAME_DEFINITION.telemetry_length, response_delay)
     length_bytes = _write_read_supcmu_i2c(i2c_master, address, f'{module_cmd_name}:TEL? {idx},LENGTH\n',
                                           SUPMCU_LENGTH_DEFINITION.telemetry_length, response_delay)
-
+    if is_simulatable_mod:
+        simulatable_bytes = _write_read_supcmu_i2c(i2c_master, address, f'{module_cmd_name}:TEL? {idx},SIMULATABLE\n',
+                                                  SUPMCU_LENGTH_DEFINITION.telemetry_length, response_delay)
     # Parse the response bytes as telemetry items, then check to see if any are not ready yet
     format_response = parse_telemetry(format_bytes, SUPMCU_FORMAT_DEFINITION)
     name_response = parse_telemetry(name_bytes, SUPMCU_NAME_DEFINITION)
     length_response = parse_telemetry(length_bytes, SUPMCU_LENGTH_DEFINITION)
-
+    if is_simulatable_mod:
+        simulatable_response = parse_telemetry(simulatable_bytes, SUPMCU_LENGTH_DEFINITION)
+        is_simulatable = bool(simulatable_response.items[0].value)
+    if is_simulatable:
+        defaults = get_values(i2c_master, address, module_cmd_name, idx, format_response.items[0].value, response_delay)
+        defaults = [item.value for item in defaults]
     # Raise exception if any of the responses are flagged as not ready.
     if not format_response.header.ready:
         raise RuntimeError(
@@ -102,24 +138,25 @@ def request_telemetry_definition(i2c_master: Union[I2CMaster, SupMCUMaster],
     if not length_response.header.ready:
         raise RuntimeError(
             f'`{module_cmd_name}:TEL? {idx},LENGTH` returned a non-ready response. try increasing `response_delay`.')
-
+    if is_simulatable_mod:
+        if not simulatable_response.header.ready:
+            raise RuntimeError(
+                f'`{module_cmd_name}:TEL? {idx},SIMULATABLE` returned a non-ready response. try increasing `response_delay`.')
     # Create the telemetry definition
     return SupMCUTelemetryDefinition(name_response.items[0].value,
                                      length_response.items[0].value,
                                      idx,
-                                     format_response.items[0].value)
+                                     format_response.items[0].value,
+                                     False if not is_simulatable_mod else simulatable_response.items[0].value,
+                                     None if not is_simulatable else defaults)
 
 
-def get_name_from_version(i2c_master: Union[I2CMaster, SupMCUMaster], address: int) -> str:
+def get_name_from_version(version: str) -> str:
     """
     Gets the Short Module name from the version string
 
-    :param i2c_master: The I2C master to write/read the requests from.
-    :param address: The address of the module on the I2C bus.
+    :param version: A version string to extract the name from
     """
-    module_def = SupMCUModuleDefinition("", "", address, {0: SupMCUTelemetryDefinition("Firmware Version", 77, 0, "S")}, {})
-    master = SupMCUMaster(i2c_master, (module_def,))
-    version = master.request_telemetry(address, TelemetryType.SupMCU, 0).items[0].value
     return version.split()[0].split("-")[0]
 
 
@@ -143,8 +180,9 @@ def request_module_definition(i2c_master: Union[I2CMaster, SupMCUMaster],
         if response_delay is None:
             response_delay = i2c_master.request_delay
         i2c_master = i2c_master.i2c_master
+    version = get_version_string(i2c_master, address)
     if module_cmd_name is None:
-        module_cmd_name = get_name_from_version(i2c_master, address)
+        module_cmd_name = get_name_from_version(version)
     if response_delay is None:
         # None was set by the SupMCU master or the passed in variable.
         response_delay = _DEFAULT_RESPONSE_DELAY
@@ -160,10 +198,12 @@ def request_module_definition(i2c_master: Union[I2CMaster, SupMCUMaster],
     amount_telemetry = parse_telemetry(amount_resp, SUPMCU_TELEMETRY_AMOUNT_DEFINITION)
     supmcu_amount, module_amount = amount_telemetry.items[0].value, amount_telemetry.items[1].value
 
+    is_simulatable_mod = "(on STM)" in version or "(on QSM)" in version
+
     supmcu_defs = {}
     for idx in range(supmcu_amount):
-        supmcu_defs[idx] = request_telemetry_definition(i2c_master, address, 'SUP', idx, response_delay)
+        supmcu_defs[idx] = request_telemetry_definition(i2c_master, address, 'SUP', idx, response_delay, is_simulatable_mod)
     module_defs = {}
     for idx in range(module_amount):
-        module_defs[idx] = request_telemetry_definition(i2c_master, address, module_cmd_name, idx, response_delay)
+        module_defs[idx] = request_telemetry_definition(i2c_master, address, module_cmd_name, idx, response_delay, is_simulatable_mod)
     return SupMCUModuleDefinition(module_name, module_cmd_name, address, supmcu_defs, module_defs)
